// main.h Header file of the CForth  project **********************************
//
// Author and (c)2020-25 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c", "CForth.c", "GForth.h",
// "dmx.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

// from cforth.h
extern uint32_t gfMSCOUNT, gfTIXCOUNT;
extern uint32_t gfTermInput;		//message from terminal iput to DEBUG();
extern uint32_t gfBackProcess;
extern uint32_t gfTibMax;
extern uint32_t gfNumBase, gfTempNumBase, gfAuto;
extern uint32_t gfRandom;
extern uint32_t gfSpiBits, gfSpiMode, gfSpiClock;
extern int32_t gfPWM1, gfPWM2, gfPWM3;
extern uint32_t gfFreq12, gfFreq3;
extern int32_t gfSTEPCOUNT1, gfSTEPCOUNT2, gfSTEPCOUNT3;
extern uint32_t gfErrors, gfParsed;
extern uint8_t gfTempBuf[2048];
#ifdef DMX
extern uint32_t gfDmxBusy;
#endif
extern uint32_t *gfSP;				//Datastack Pointer
extern uint32_t *gfRP;		        //Returnstack Pointer
extern uint32_t *gfIP;		        //Instruction Pointer
extern uint32_t gfTCode[];

extern struct USERHEADER
{
uint8_t name[16];
uint32_t toCompile;
uint32_t attributes;
} gfUserHeader[MAXUSER];
//6 words = 24 bytes/struct * 168 entries = 1008 words = 4032 bytes)

extern void quit(void);
extern uint32_t query(void);
extern void eval(void);
extern uint32_t interpretUserToken(void);
extern uint32_t executeTCode(void);
extern void emptyForth(void);
extern void subFREQ12();
extern void subPWM1();
extern void subPWM2();
extern void subFREQ3(void);
extern void subPWM3(void);
extern void op_BAKOP(void);
extern void op_FLUSH(void);
extern void op_ABORT(void);
extern void op_QUEST(void);
extern void op_DMXLOAD(void);

// from usb-fs.c  ************
extern uint32_t guDeviceClass, guConfigured;
extern void stopUSB(void);
extern void initUSB(void);
extern uint16_t pullUsbRxBuf(void);
extern void sendUsbByte(uint8_t);
extern void enCodeMidi(uint8_t);

#ifdef DMX
// from dmx.c
extern uint8_t gdDMXBUF[DMX_BUFLEN+2];
extern uint8_t gdFINALMEM[DMX_BUFLEN+2];
extern uint32_t gdDmxAddr, gdDmxTxIndex, gdDmxMax;
extern uint32_t gdFaderBusy;
extern uint32_t gdFadeTime, gdBootScene;
extern void startDmxTx(void);
extern void DmxHeader(void);
extern void fadeProcess(void);
extern void startFader(uint32_t);
#endif

// ****************************************

//buffer length defined in "globaldefines.h"
uint8_t gUSART_RXBUF[USART_RXBUFLEN+4];
uint8_t *gUSART_RXHEAD=0,*gUSART_RXTAIL=0,*gUSART_RXBUFTOP= &(gUSART_RXBUF[USART_RXBUFLEN]);
#define USART_TXBUFLEN 4000
uint8_t gUSART_TXBUF[USART_TXBUFLEN+4];
uint8_t *gUSART_TXHEAD=0,*gUSART_TXTAIL=0,*gUSART_TXBUFTOP= &(gUSART_TXBUF[USART_TXBUFLEN]);
//system control
uint32_t gSYSCOUNT, gLEDCOUNT, gDELAYCOUNT, gTIMEOUT;
uint32_t gProject, gLedBusy;

#define PAGE23 (0x0800B800)         //starting from 46kB Flash
#define PAGE24 (0x0800C000)         //starting from 48kB Flash
#define PAGE25 (0x0800C800)
#define PAGE26 (0x0800D000)
#define PAGE27 (0x0800D800)
#define PAGE28 (0x0800E000)
#define PAGE29 (0x0800E800)
#define PAGE30 (0x0800F000)
#define PAGE31 (0x0800F800)         //16kB totally
//project 2 from 64kB: pages 32,33 param+UserToken, 34,35,36,37,38,39 code
//project 3:from 80kB: pages 40,41 param+UserToken, 42,43,44,45,46,47 code
//project 4:from 96kB: pages 48,49 param+UserToken, 50,51,52,53,54,55 code
//project 5:from 112kB:pages 56,57 param+UserToken, 58,59,60,61,62,63 code
//if option DMX:
#define PAGE40 (0x08014000)	//pages 40..63 reserved for DMX, 48kB totally

void sendModel(void);
void SetSysClock(void);
void LED_Off(void);
void msDelay(uint32_t);
void wait5us(void);
void popFlashPage23(void);
uint32_t pushFlashPage23(void);
void popFlashProject();
uint32_t pushFlashProject(void);
void servSystem(void);
void startUSB(void);
//USART
void initUsart2(void);

void initUsart1(void);
void sendDebug(uint8_t);

void sendB(uint8_t);
void sendTB(uint8_t);
void sendIOB(uint8_t);
void sendNibble(uint8_t);
void sendString(const uint8_t *);
void sendX8(uint32_t);
void sendX32(uint32_t);
void sendNX32(uint32_t);
void send0X32(uint32_t);
void sendLS32(uint32_t);
void sendS32(uint32_t);
void sendNS32(uint32_t);
uint32_t pullUsartRxBuf(void);
uint32_t pullTermRxBuf(void);
uint8_t parseHexNibble(uint8_t);
uint32_t parseHexNum(uint8_t);
uint8_t parseDeciDigit(void);
void MidiScanner(uint32_t);
uint32_t Midi2Forth(uint32_t, uint32_t);
//Interrupt handlers
//void SysTick_Handler(void);
//void USART1_IRQHandler(void);
